@@ -6,9 +6,10 @@ import VITAE.model as model
 import VITAE.preprocess as preprocess
 import VITAE.train as train
 from VITAE.inference import Inferer
-from VITAE.utils import load_data, get_embedding, get_igraph, louvain_igraph, \
-    plot_clusters, plot_marker_gene, DE_test
+from VITAE.utils import load_data, get_embedding, get_igraph, leidenalg_igraph, \
+    plot_clusters, plot_marker_gene, plot_uncertainty, DE_test
 from VITAE.metric import topology, get_GRI
+import tensorflow as tf
 
 from sklearn.metrics.cluster import adjusted_rand_score
 from sklearn.model_selection import train_test_split
@@ -170,11 +171,11 @@ class VITAE():
             np.save(f, np.array([
                 self.dim_origin, self.dimensions, self.dim_latent,
                 self.data_type, False if self.c_score is None else True], dtype=object))
-        if hasattr(self, 'pi'):
+        if hasattr(self, 'inferer') and hasattr(self.inferer, 'embed_mu'):
             with open(path_to_file+'.inference', 'wb') as f:
                 np.save(f, np.array([
                     self.pi, self.mu, self.pc_x, self.w_tilde, self.var_w_tilde,
-                    self.D_JS, self.z, self.embed_z, self.inferer.embed_mu], dtype=object))
+                    self.z, self.embed_z, self.inferer.embed_mu], dtype=object))
     
 
     def load_model(self, path_to_file: str = 'model.checkpoint', load_labels: bool = False):
@@ -209,8 +210,13 @@ class VITAE():
             self.init_latent_space(n_clusters, cluster_labels)
             if os.path.exists(path_to_file+'.inference'):
                 with open(path_to_file+'.inference', 'rb') as f:
-                    [self.pi, self.mu, self.pc_x, self.w_tilde, self.var_w_tilde,
-                        self.D_JS, self.z, self.embed_z, embed_mu] = np.load(f, allow_pickle=True)
+                    arr = np.load(f, allow_pickle=True)
+                    if len(arr)==9:
+                        [self.pi, self.mu, self.pc_x, self.w_tilde, self.var_w_tilde,
+                            self.D_JS, self.z, self.embed_z, embed_mu] = arr
+                    else:
+                        [self.pi, self.mu, self.pc_x, self.w_tilde, self.var_w_tilde,
+                            self.z, self.embed_z, embed_mu] = arr
                 self.inferer.mu = self.mu
                 self.inferer.embed_z = self.embed_z
                 self.inferer.embed_mu = embed_mu
@@ -219,9 +225,9 @@ class VITAE():
 
 
     def pre_train(self, stratify = False, test_size = 0.1, random_state: int = 0,
-            learning_rate: float = 1e-3, batch_size: int = 32, L: int = 1, alpha: float = 0.01,
+            learning_rate: float = 1e-3, batch_size: int = 32, L: int = 1, alpha: float = 0.10,
             num_epoch: int = 300, num_step_per_epoch: Optional[int] = None,
-            early_stopping_patience: int = 10, early_stopping_tolerance: float = 1e-3, early_stopping_warmup: int = 0, 
+            early_stopping_patience: int = 5, early_stopping_tolerance: float = 1.0,
             path_to_weights: Optional[str] = None):
         '''Pretrain the model with specified learning rate.
 
@@ -249,8 +255,6 @@ class VITAE():
             The maximum number of epoches if there is no improvement.
         early_stopping_tolerance : float, optional 
             The minimum change of loss to be considered as an improvement.
-        early_stopping_warmup : int, optional
-            The number of warmup epoches.
         path_to_weights : str, optional 
             The path of weight file to be saved; not saving weight if None.
         '''                    
@@ -265,16 +269,16 @@ class VITAE():
                                 random_state=random_state)
         if num_step_per_epoch is None:
             num_step_per_epoch = len(id_train)//batch_size+1
-        self.train_dataset = train.warp_dataset(self.X_normalized[id_train], 
-                                                None if self.c_score is None else self.c_score[id_train],
+        self.train_dataset = train.warp_dataset(self.X_normalized[id_train].astype(tf.keras.backend.floatx()), 
+                                                None if self.c_score is None else self.c_score[id_train].astype(tf.keras.backend.floatx()),
                                                 batch_size, 
-                                                self.X[id_train], 
-                                                self.scale_factor[id_train])
+                                                self.X[id_train].astype(tf.keras.backend.floatx()), 
+                                                self.scale_factor[id_train].astype(tf.keras.backend.floatx()))
         self.test_dataset = train.warp_dataset(self.X_normalized[id_test], 
-                                                None if self.c_score is None else self.c_score[id_test],
+                                                None if self.c_score is None else self.c_score[id_test].astype(tf.keras.backend.floatx()),
                                                 batch_size, 
-                                                self.X[id_test], 
-                                                self.scale_factor[id_test])
+                                                self.X[id_test].astype(tf.keras.backend.floatx()), 
+                                                self.scale_factor[id_test].astype(tf.keras.backend.floatx()))
         self.vae = train.pre_train(
             self.train_dataset,
             self.test_dataset,
@@ -285,7 +289,7 @@ class VITAE():
             num_step_per_epoch,
             early_stopping_patience,
             early_stopping_tolerance,
-            early_stopping_warmup)
+            0)
 
         if path_to_weights is not None:
             self.save_model(path_to_weights)
@@ -306,6 +310,8 @@ class VITAE():
     def set_cell_subset(self, selected_cell_names):
         '''Set a subset of interested cells.
 
+        This will have influence on the downstream processes (pretraining, training, inference, DE test, ...).
+
         Parameters
         ----------
         selected_cell_names : np.array, optional
@@ -313,7 +319,8 @@ class VITAE():
         ''' 
         self.selected_cell_subset = np.unique(selected_cell_names)
         self.selected_cell_subset_id = np.sort(np.where(np.in1d(self.cell_names, selected_cell_names))[0])
-        
+        print("Setting cell subset with length %d."%len(self.selected_cell_subset))
+
     
     def refine_pi(self, batch_size: int = 64):  
         '''Refine pi by the its posterior. This function will be effected if 'selected_cell_subset_id' is set.
@@ -373,9 +380,9 @@ class VITAE():
 
     def train(self, stratify = False, test_size = 0.1, random_state: int = 0,
             learning_rate: float = 1e-3, batch_size: int = 32, 
-            L: int = 1, alpha: float = 0.01, beta: float = 1, 
+            L: int = 1, alpha: float = 0.10, beta: float = 1, 
             num_epoch: int = 300, num_step_per_epoch: Optional[int] =  None,
-            early_stopping_patience: int = 10, early_stopping_tolerance: float = 1e-3, early_stopping_warmup: int = 5,
+            early_stopping_patience: int = 5, early_stopping_tolerance: float = 1.0, early_stopping_warmup: int = 10,
             path_to_weights: Optional[str] = None, plot_every_num_epoch: Optional[int] = None, dimred: str = 'umap', **kwargs):
         '''Train the model.
 
@@ -427,21 +434,21 @@ class VITAE():
                                 random_state=random_state)
         if num_step_per_epoch is None:
             num_step_per_epoch = len(id_train)//batch_size+1
-        c = None if self.c_score is None else self.c_score[self.selected_cell_subset_id,:]
-        self.train_dataset = train.warp_dataset(self.X_normalized[self.selected_cell_subset_id,:][id_train],
+        c = None if self.c_score is None else self.c_score[self.selected_cell_subset_id,:].astype(tf.keras.backend.floatx())
+        self.train_dataset = train.warp_dataset(self.X_normalized[self.selected_cell_subset_id,:][id_train].astype(tf.keras.backend.floatx()),
                                                 None if c is None else c[id_train],
                                                 batch_size, 
-                                                self.X[self.selected_cell_subset_id,:][id_train], 
-                                                self.scale_factor[self.selected_cell_subset_id][id_train])
-        self.test_dataset = train.warp_dataset(self.X_normalized[self.selected_cell_subset_id,:][id_test],
+                                                self.X[self.selected_cell_subset_id,:][id_train].astype(tf.keras.backend.floatx()), 
+                                                self.scale_factor[self.selected_cell_subset_id][id_train].astype(tf.keras.backend.floatx()))
+        self.test_dataset = train.warp_dataset(self.X_normalized[self.selected_cell_subset_id,:][id_test].astype(tf.keras.backend.floatx()),
                                                 None if c is None else c[id_test],
                                                 batch_size, 
-                                                self.X[self.selected_cell_subset_id,:][id_test], 
-                                                self.scale_factor[self.selected_cell_subset_id][id_test])    
+                                                self.X[self.selected_cell_subset_id,:][id_test].astype(tf.keras.backend.floatx()), 
+                                                self.scale_factor[self.selected_cell_subset_id][id_test].astype(tf.keras.backend.floatx()))    
         if plot_every_num_epoch is None:
             self.whole_dataset = None    
         else:
-            self.whole_dataset = train.warp_dataset(self.X_normalized[self.selected_cell_subset_id,:], 
+            self.whole_dataset = train.warp_dataset(self.X_normalized[self.selected_cell_subset_id,:].astype(tf.keras.backend.floatx()), 
                                                     c,
                                                     batch_size)                                    
         self.vae = train.train(
@@ -460,7 +467,7 @@ class VITAE():
             early_stopping_warmup,            
             self.labels[self.selected_cell_subset_id],            
             plot_every_num_epoch,
-            dimred='umap', 
+            dimred, 
             **kwargs            
             )
             
@@ -470,7 +477,7 @@ class VITAE():
 
     def init_inference(self, batch_size: int = 32, L: int = 5, 
             dimred: str = 'umap', refit_dimred: bool = True, **kwargs):
-        '''Initialze trajectory inference by computing the posterior estimations.        
+        '''Initialize trajectory inference by computing the posterior estimations.        
 
         Parameters
         ----------
@@ -485,21 +492,22 @@ class VITAE():
         **kwargs :  
             Extra key-value arguments for dimension reduction algorithms.              
         '''
-        c = None if self.c_score is None else self.c_score[self.selected_cell_subset_id,:]
-        self.test_dataset = train.warp_dataset(self.X_normalized[self.selected_cell_subset_id,:], 
+        c = None if self.c_score is None else self.c_score.astype(tf.keras.backend.floatx())
+        self.test_dataset = train.warp_dataset(self.X_normalized.astype(tf.keras.backend.floatx()), 
                                                c,
                                                batch_size)
         self.pi, self.mu, self.pc_x,\
-            self.w_tilde,self.var_w_tilde,self.D_JS,self.z = self.vae.inference(self.test_dataset, L=L)
+            self.w_tilde,self.var_w_tilde,self.z = self.vae.inference(self.test_dataset, L=L)
         if refit_dimred or not hasattr(self.inferer, 'embed_z'):
-            self.embed_z = self.inferer.init_embedding(self.z, self.mu, **kwargs)
+            print('Computing embeddings.')
+            self.embed_z = self.inferer.init_embedding(self.z, self.mu, dimred, **kwargs)
         else:
             self.embed_z = self.inferer.embed_z
         return None
         
 
     def select_root(self, days, method: str = 'sum'):
-        '''Initialze trajectory inference by computing the posterior estimations.        
+        '''Select the root vertex based on days information.      
 
         Parameters
         ----------
@@ -557,7 +565,9 @@ class VITAE():
         G : nx.Graph 
             The weighted graph with weight on each edge indicating its score of existence.
         '''
-        G, edges = self.inferer.init_inference(self.w_tilde, self.pc_x, thres, method, no_loop)
+        G, _ = self.inferer.init_inference(self.w_tilde[self.selected_cell_subset_id,:], 
+                                                self.pc_x[self.selected_cell_subset_id,:], 
+                                                thres, method, no_loop)
         if is_plot:
             self.inferer.plot_clusters(self.cluster_labels, plot_labels=plot_labels, path=path)
         return G
@@ -579,20 +589,27 @@ class VITAE():
 
         Returns
         ----------
-        G : nx.Graph 
+        modified_G : nx.Graph 
             The modified graph that indicates the inferred trajectory.
-        w : np.array
+        modified_w_tilde : np.array
             \([N,k]\) The modified \(\\tilde{w}\).
         pseudotime : np.array
             \([N,]\) The pseudotime based on projected trajectory.
         '''
-        G, w, pseudotime = self.inferer.infer_trajectory(init_node, 
+        self.modified_G, modified_w_tilde, pseudotime = self.inferer.infer_trajectory(init_node, 
                                                          self.label_names[self.selected_cell_subset_id], 
                                                          cutoff, 
                                                          path=path, 
                                                          is_plot=is_plot)
-        self.pseudotime = pseudotime
-        return G, w, pseudotime
+        if len(self.selected_cell_subset_id)<len(self.cell_names):
+            self.modified_w_tilde = np.full(self.w_tilde.shape, np.nan)
+            self.modified_w_tilde[self.selected_cell_subset_id,:] = modified_w_tilde
+            self.pseudotime = np.full(len(self.cell_names), np.nan)
+            self.pseudotime[self.selected_cell_subset_id] = pseudotime 
+        else:
+            self.modified_w_tilde = modified_w_tilde
+            self.pseudotime = pseudotime 
+        return self.modified_G, modified_w_tilde, pseudotime
 
 
     def differentially_expressed_test(self, alpha: float = 0.05):
@@ -626,6 +643,35 @@ class VITAE():
         return res_df
 
 
+    def plot_uncertainty(self, refit_dimred: bool = False, dimred: str = 'umap', path: Optional[str] =None, **kwargs):
+        '''Plot uncertainty of all selected cells.
+
+        Parameters
+        ----------        
+        refit_dimred : boolean, optional 
+            Whether to refit dimension reduction or use the existing embedding after inference.
+        dimred : str, optional
+            The name of dimension reduction algorithms, can be 'umap', 'pca' and 'tsne'.
+        path : str, optional
+            The path to save the figure, or not saving if it is None.
+        **kwargs :  
+            Extra key-value arguments for dimension reduction algorithms.
+        '''
+        if not hasattr(self, 'modified_w_tilde'):
+            raise ReferenceError("The object 'modified_w_tilde' does not exist! Please infer a trajectory before calling this function.")
+        
+        uncertainty = np.sum((self.modified_w_tilde - self.w_tilde)**2, axis=-1) + np.sum(self.var_w_tilde, axis=-1)
+
+        if not hasattr(self, 'embed_z') or refit_dimred:
+            z = self.get_latent_z()       
+            embed_z = get_embedding(z, dimred, **kwargs)
+        else:
+            embed_z = self.embed_z
+        return plot_uncertainty(uncertainty[self.selected_cell_subset_id], 
+                                embed_z[self.selected_cell_subset_id,:],
+                                path)
+
+
     def plot_marker_gene(self, gene_name: str, refit_dimred: bool = False, dimred: str = 'umap', path: Optional[str] =None, **kwargs):
         '''Plot expression of the given marker gene.
 
@@ -653,11 +699,10 @@ class VITAE():
             embed_z = get_embedding(z, dimred, **kwargs)
         else:
             embed_z = self.embed_z
-        plot_marker_gene(expression, 
-                         gene_name, 
-                         embed_z[self.selected_cell_subset_id,:],
-                         path)
-        return None
+        return plot_marker_gene(expression, 
+                                gene_name, 
+                                embed_z[self.selected_cell_subset_id,:],
+                                path)
 
 
     def evaluate(self, milestone_net, begin_node_true, grouping = None,
