@@ -1,12 +1,12 @@
 __all__ = ['VITAE']
 
-from VITAE.VITAE import VITAE, get_embedding, get_igraph, louvain_igraph, plot_clusters, load_data
+from VITAE.VITAE import VITAE, get_embedding, get_igraph, leidenalg_igraph, plot_clusters, load_data
 
 __author__ = "Jin-Hong Du, Ming Gao, and Jingshu Wang"
 __copyright__ = "Copyright 2020, The Trajectory Inference Project"
 
 __license__ = "MIT"
-__version__ = "1.1.1"
+__version__ = "1.1.5"
 __maintainer__ = "Jin-Hong Du"
 __email__ = "dujinhong@uchicago.edu"
 __status__ = "Development"
\ No newline at end of file
