@@ -180,6 +180,7 @@ class Inferer(object):
         self.mu = mu.copy()
         concate_z = np.concatenate((z, mu.T), axis=0)
         embed = get_embedding(concate_z, dimred, **kwargs)
+        embed = (embed - np.min(embed, axis=0, keepdims=True)) / (np.max(embed, axis=0, keepdims=True) - np.min(embed, axis=0, keepdims=True))
         
         self.embed_z = embed[:-self.NUM_CLUSTER,:]
         self.embed_mu = embed[-self.NUM_CLUSTER:,:]
@@ -206,7 +207,7 @@ class Inferer(object):
             fig, ax = plt.subplots(1,1, figsize=(20, 10))
             for i,x in enumerate(np.unique(labels)):
                 ax.scatter(*self.embed_z[labels==x].T, c=[colors[i]],
-                    s=8, alpha=0.6, label=str(x))
+                    s=16, alpha=0.5, label=str(x))
                 if plot_labels:
                     ax.text(np.mean(self.embed_z[labels==x,0]), 
                             np.mean(self.embed_z[labels==x,1]), str(x), fontsize=16)
@@ -214,7 +215,7 @@ class Inferer(object):
             ax.set_position([box.x0, box.y0 + box.height * 0.1,
                              box.width, box.height * 0.9])
             ax.legend(loc='upper center', bbox_to_anchor=(0.5, -0.05),
-                      fancybox=True, shadow=True, markerscale=5, ncol=5)
+                      fancybox=True, shadow=True, markerscale=3, ncol=5)
             ax.set_title('Cluster Membership')
             plt.setp(ax, xticks=[], yticks=[])
             if path is not None:
@@ -327,7 +328,7 @@ class Inferer(object):
         '''
         # select edges
         if len(self.edges)==0:
-            select_edges = []
+            milestone_net = select_edges = []
             G = nx.Graph()
             G.add_nodes_from(self.G.nodes)
         else:
@@ -344,6 +345,8 @@ class Inferer(object):
             subG = G.subgraph(connected_comps)
             if len(subG.edges)>0:                
                 milestone_net = self.build_milestone_net(subG,init_node)
+                if self.no_loop is False and milestone_net.shape[0]<len(G.edges):
+                    warnings.warn("The directed graph shown is a minimum spanning tree of the estimated trajectory backbone to avoid arbitrary assignment of the directions.")
                 select_edges = milestone_net[:,:2]
                 select_edges_score = graph[select_edges[:,0], select_edges[:,1]]
                 if select_edges_score.max() - select_edges_score.min() == 0:
@@ -369,14 +372,14 @@ class Inferer(object):
                 sc = ax.scatter(*self.embed_z[pseudotime>-1,:].T,
                     norm=norm,
                     c=pseudotime[pseudotime>-1],
-                    s=8, alpha=0.5)
+                    s=16, alpha=0.5)
                 plt.colorbar(sc, ax=[ax], location='right')
             else:
                 norm = None
                 
             if np.sum(pseudotime==-1)>0:
                 ax.scatter(*self.embed_z[pseudotime==-1,:].T,
-                            c='gray', s=8, alpha=0.4)
+                            c='gray', s=16, alpha=0.4)
             
             for i in range(len(select_edges)):
                 points = self.embed_z[np.sum(w[:,select_edges[i,:]]>0, axis=-1)==2,:]
@@ -400,14 +403,14 @@ class Inferer(object):
 
                 delta_x = self.embed_mu[select_edges[i,1], 0]-x_smooth[-2]
                 delta_y = self.embed_mu[select_edges[i,1], 1]-y_smooth[-2]
-                length = np.sqrt(delta_x**2 + delta_y**2) * 1.5                
+                length = np.sqrt(delta_x**2 + delta_y**2) * 50              
                 ax.arrow(
                         self.embed_mu[select_edges[i,1], 0]-delta_x/length, 
                         self.embed_mu[select_edges[i,1], 1]-delta_y/length, 
                         delta_x/length,
                         delta_y/length,
                         color='black', alpha=1.0,
-                        shape='full', lw=0, length_includes_head=True, head_width=0.4, zorder=2)
+                        shape='full', lw=0, length_includes_head=True, head_width=0.02, zorder=2)
             
             for i in range(len(self.CLUSTER_CENTER)):
                 ax.scatter(*self.embed_mu[i:i+1,:].T, c=[colors[i]],
