@@ -41,7 +41,7 @@ def warp_dataset(X_normalized, c_score, batch_size:int, X=None, scale_factor=Non
     '''
     # fake c_score
     if c_score is None:
-        c_score = np.zeros((X_normalized.shape[0],1), np.float32)
+        c_score = np.zeros((X_normalized.shape[0],1), tf.keras.backend.floatx())
         
     if X is not None:
         train_dataset = tf.data.Dataset.from_tensor_slices((X, X_normalized, c_score, scale_factor))
@@ -179,13 +179,14 @@ def train(train_dataset, test_dataset, whole_dataset, vae,
     vae : VariationalAutoEncoder
         The trained model.
     '''   
+    optimizer_ = tf.keras.optimizers.Adam(learning_rate)
     optimizer = tf.keras.optimizers.Adam(learning_rate)
     loss_test = [tf.keras.metrics.Mean() for _ in range(4)]
     loss_train = [tf.keras.metrics.Mean() for _ in range(4)]
     early_stopping = Early_Stopping(patience = early_stopping_patience, tolerance = early_stopping_tolerance, warmup=early_stopping_warmup)
 
     print('Warmup:%d'%early_stopping_warmup)
-    weight = np.array([1,beta,beta], dtype=np.float32)
+    weight = np.array([1,beta,beta], dtype=tf.keras.backend.floatx())
     weight = tf.convert_to_tensor(weight)
     
     for epoch in range(num_epoch):
@@ -201,7 +202,7 @@ def train(train_dataset, test_dataset, whole_dataset, vae,
                     loss = tf.reduce_sum(losses[1:])
                 grads = tape.gradient(loss, vae.latent_space.trainable_weights,
                             unconnected_gradients=tf.UnconnectedGradients.ZERO)
-                optimizer.apply_gradients(zip(grads, vae.latent_space.trainable_weights))
+                optimizer_.apply_gradients(zip(grads, vae.latent_space.trainable_weights))
             else:
                 with tf.GradientTape() as tape:
                     losses = vae(x_norm_batch, c_score, x_batch, x_scale_factor, L=L, alpha=alpha)
@@ -250,7 +251,7 @@ def train(train_dataset, test_dataset, whole_dataset, vae,
 
 
         if plot_every_num_epoch is not None and (epoch%plot_every_num_epoch==0 or epoch==num_epoch-1):
-            _, mu, _, w_tilde, _, _, z_mean = vae.inference(whole_dataset, 1)
+            _, mu, _, w_tilde, _, z_mean = vae.inference(whole_dataset, 1)
             c = np.argmax(w_tilde, axis=-1)
             
             concate_z = np.concatenate((z_mean, mu.T), axis=0)
